module kinddefs

  implicit none
  private
! single precision (IEEE 754)
  integer, parameter :: system_r4=selected_real_kind(6, 37)
! double precision (IEEE 754)
  integer, parameter :: system_r8=selected_real_kind(15, 307)
  integer, parameter, public :: dp =system_r8 ! default precision
  integer, parameter, public :: jp =system_r8 ! Diagonal LU Jacobian precision
  integer, parameter, public :: odp=system_r4 ! Off-diagonal Jacobian precision

end module kinddefs


module data_module

  use kinddefs, only : dp, jp, odp

  implicit none

  private

  public :: load_data, viscosity_law, color_cells

  integer, public :: ivgrd, max_node_per_cell, max_face_per_cell
  integer, public :: nnodes0, nnodes01, ncell, nnz0, face_per_cell
  integer, public :: node_per_cell, edge_per_cell, nnz01, n_tot, njac
  integer, public :: ncell_colors

  integer, dimension(:),   allocatable, public :: ia, ja, nzg2m, g2m
  integer, dimension(:),   allocatable, public :: ncell_in_color
  integer, dimension(:,:), allocatable, public :: c2n, local_f2n, local_e2n
  integer, dimension(:,:), allocatable, public :: chk_norm

  real(dp), public :: tref, xmach, re, gamma, gm1, ggm1, sutherland_constant
  real(dp), public :: prandtl, turbulent_prandtl, golden

  real(dp),  dimension(:),     allocatable, public :: x, y, z, amut
  real(dp),  dimension(:,:),   allocatable, public :: qnode
  real(jp),  dimension(:,:,:), allocatable, public :: a_diag
  real(odp), dimension(:,:,:), allocatable, public :: a_off

  character(len=80), public :: datafile, color_flag

  logical, parameter, public :: timing = .true.

contains

  subroutine load_data()

    write(*,*) 'Loading data...'

    open(66,file=trim(datafile),form='unformatted',status='old')

    read(66) ivgrd, max_node_per_cell, max_face_per_cell, nnodes0, nnodes01,   &
             ncell, nnz0, face_per_cell, node_per_cell, edge_per_cell, nnz01,  &
             n_tot, njac, tref, xmach, re, gamma, gm1, ggm1,                   &
             sutherland_constant, prandtl, turbulent_prandtl

    write(*,*) 'Number of points   in grid: ', nnodes0
    write(*,*) 'Number of elements in grid: ', ncell

    allocate(ia(nnodes01+1))
    read(66) ia

    allocate(ja(nnz01))
    read(66) ja

    allocate(nzg2m(nnz01))
    read(66) nzg2m

    allocate(g2m(nnodes01))
    read(66) g2m

    allocate(c2n(node_per_cell,ncell))
    read(66) c2n

    allocate(local_f2n(face_per_cell,4))
    read(66) local_f2n

    allocate(local_e2n(edge_per_cell,6))
    read(66) local_e2n

    allocate(chk_norm(face_per_cell,face_per_cell))
    read(66) chk_norm

    allocate(x(nnodes01))
    read(66) x

    allocate(y(nnodes01))
    read(66) y

    allocate(z(nnodes01))
    read(66) z

    allocate(amut(nnodes01))
    read(66) amut

    allocate(qnode(n_tot,nnodes01))
    read(66) qnode

    allocate(a_diag(njac,njac,nnodes0))
    a_diag = 0.0_dp

    allocate(a_off(njac,njac,nnz0))
    a_off = 0.0_odp

    read(66) golden

    close(66)

    write(*,*) 'Done loading data...'

  end subroutine load_data

  pure function viscosity_law( cstar, t )

    use kinddefs, only : dp

    real(dp), intent(in) :: cstar, t
    real(dp)             :: viscosity_law

  continue

    viscosity_law = (1.0_dp+cstar)/(t+cstar)*t*sqrt(t)

  end function viscosity_law


  subroutine color_cells()

    integer :: i, j, min_color, ncell_remaining, new_cells, inode

    integer, dimension(ncell)    :: cell_color
    integer, dimension(nnodes01) :: node_tag

    integer, dimension(node_per_cell,ncell) :: temp_c2n

    integer, dimension(:), allocatable :: temp_ncell_in_color

  continue

    allocate(ncell_in_color(1))

    if ( trim(color_flag) == '0' ) then
      ncell_colors = 1
      ncell_in_color(1) = ncell
      return
    endif

    ncell_colors    = 0
    cell_color(:)   = 0
    ncell_remaining = ncell

    coloring : do

      if ( ncell_remaining == 0 ) exit coloring

      ncell_colors = ncell_colors + 1

      allocate(temp_ncell_in_color(ncell_colors))
      temp_ncell_in_color(1:ncell_colors-1) = ncell_in_color(1:ncell_colors-1)
      deallocate(ncell_in_color)
      call move_alloc(temp_ncell_in_color,ncell_in_color)

      new_cells = 0

      node_tag(:) = 0

      search_cells : do i = 1, ncell

        if ( cell_color(i) /= 0 ) cycle search_cells

        do j = 1, node_per_cell
          inode = c2n(j,i)
          if ( node_tag(inode) /= 0 ) cycle search_cells
        end do

        cell_color(i) = ncell_colors

        do j = 1, node_per_cell
          inode = c2n(j,i)
          node_tag(inode) = 1
        end do

        new_cells = new_cells + 1
        ncell_remaining = ncell_remaining - 1

      end do search_cells

      if ( new_cells == 0 .and. ncell_remaining > 0 ) then
        write(*,*) 'Error in finding cell in color...'
        stop
      endif

      ncell_in_color(ncell_colors) = new_cells

    end do coloring

! Make sure everyone got colored

    min_color = minval(cell_color)

    if ( min_color /= 1 ) then
      write(*,*) 'Cell(s) not colored in color_cells...'
      stop
    endif

    if ( sum(ncell_in_color) /= ncell ) then
      write(*,*) 'Incorrect number of colored cells...'
      stop
    endif

    new_cells = 0

    do i = 1, ncell_colors
      do j = 1, ncell
        if ( cell_color(j) == i ) then
          new_cells = new_cells + 1
          temp_c2n(:,new_cells) = c2n(:,j)
        endif
      end do
    end do

    c2n(:,:) = temp_c2n(:,:)

  end subroutine color_cells

end module data_module


module jacobian_module

  implicit none
  private
  public :: edgejp_mix_cell_opt, edgejp_done

contains

  subroutine edgejp_mix_cell_opt()

    use kinddefs,    only : dp, odp
    use data_module, only : tref, xmach, re, ivgrd, gamma, gm1, ggm1,          &
                            sutherland_constant, prandtl, turbulent_prandtl,   &
                            max_node_per_cell, max_face_per_cell,              &
                            viscosity_law, nnodes0, nnodes01, ncell, nnz0, c2n,&
                            x, y, z, qnode, amut, a_diag, a_off, local_f2n,    &
                            local_e2n, chk_norm, face_per_cell, node_per_cell, &
                            edge_per_cell, nnz01, ia, ja, n_tot, njac, nzg2m,  &
                            g2m, ncell_colors, ncell_in_color

    integer     :: column, n, nodec, j, k, ioff, row
    integer     :: ie, i, icolor, nstart
    integer     :: n1_loc, n2_loc, node
    integer     :: big_angle, n3_loc, n4_loc, n5_loc, n6_loc
    integer     :: iface, nn1, nn2, nn3, nn4

    real(dp)    :: c23, c43, cgp, cgpt, xmr
    real(dp)    :: cstar, dot, disi, disiex, disiey, disiez
    real(dp)    :: ex, ey, ez
    real(dp)    :: xr, yr, zr, xl, yl, zl, xm, ym, zm
    real(dp)    :: areax, areay, areaz
    real(dp)    :: xc, yc, zc, cell_vol, fact
    real(dp)    :: rmu, umu, vmu, wmu, rmucgp
    real(dp)    :: dtxxdr, dtxxdm, dtxxdn, dtxxdl
    real(dp)    :: dtxydr, dtxydm, dtxydn
    real(dp)    :: dtxzdr, dtxzdm, dtxzdl
    real(dp)    :: dtyydr, dtyydm, dtyydn, dtyydl
    real(dp)    :: dtyzdr, dtyzdn, dtyzdl
    real(dp)    :: dtzzdr, dtzzdm, dtzzdn, dtzzdl
    real(dp)    :: rax, ray, raz, tax, tay, taz
    real(dp)    :: rmuinv
    real(dp)    :: xavg,yavg,zavg
    real(dp)    :: xavg1,yavg1,zavg1,xavg2,yavg2,zavg2
    real(dp)    :: nx1,ny1,nz1,nx2,ny2,nz2,term1,term2
    real(dp)    :: cell_vol_inv

    real(dp)                                    :: dgradu_xidr, dgradu_xidm
    real(dp), dimension(max_node_per_cell)      :: duxavgdr
    real(dp), dimension(max_node_per_cell)      :: duxavgdm
    real(dp), dimension(max_node_per_cell)      :: duyavgdr
    real(dp), dimension(max_node_per_cell)      :: duyavgdm
    real(dp), dimension(max_node_per_cell)      :: duzavgdr
    real(dp), dimension(max_node_per_cell)      :: duzavgdm

    real(dp)                                    :: dgradv_xidr, dgradv_xidn
    real(dp), dimension(max_node_per_cell)      :: dvxavgdr
    real(dp), dimension(max_node_per_cell)      :: dvxavgdn
    real(dp), dimension(max_node_per_cell)      :: dvyavgdr
    real(dp), dimension(max_node_per_cell)      :: dvyavgdn
    real(dp), dimension(max_node_per_cell)      :: dvzavgdr
    real(dp), dimension(max_node_per_cell)      :: dvzavgdn

    real(dp)                                    :: dgradw_xidr, dgradw_xidl
    real(dp), dimension(max_node_per_cell)      :: dwxavgdr
    real(dp), dimension(max_node_per_cell)      :: dwxavgdl
    real(dp), dimension(max_node_per_cell)      :: dwyavgdr
    real(dp), dimension(max_node_per_cell)      :: dwyavgdl
    real(dp), dimension(max_node_per_cell)      :: dwzavgdr
    real(dp), dimension(max_node_per_cell)      :: dwzavgdl

    real(dp)                                    :: dgradt_xidr, dgradt_xidm
    real(dp)                                    :: dgradt_xidn, dgradt_xidl
    real(dp)                                    :: dgradt_xide
    real(dp)                                    :: rhoinv
    real(dp), dimension(max_node_per_cell)      :: dtxavgdr
    real(dp), dimension(max_node_per_cell)      :: dtxavgdm
    real(dp), dimension(max_node_per_cell)      :: dtxavgdn
    real(dp), dimension(max_node_per_cell)      :: dtxavgdl
    real(dp), dimension(max_node_per_cell)      :: dtxavgde
    real(dp), dimension(max_node_per_cell)      :: dtyavgdr
    real(dp), dimension(max_node_per_cell)      :: dtyavgdm
    real(dp), dimension(max_node_per_cell)      :: dtyavgdn
    real(dp), dimension(max_node_per_cell)      :: dtyavgdl
    real(dp), dimension(max_node_per_cell)      :: dtyavgde
    real(dp), dimension(max_node_per_cell)      :: dtzavgdr
    real(dp), dimension(max_node_per_cell)      :: dtzavgdm
    real(dp), dimension(max_node_per_cell)      :: dtzavgdn
    real(dp), dimension(max_node_per_cell)      :: dtzavgdl
    real(dp), dimension(max_node_per_cell)      :: dtzavgde

    real(dp), dimension(max_node_per_cell)      :: dtdr,dtdm,dtdn,dtdl,dtde
    real(dp), dimension(max_node_per_cell)      :: duxdr, duxdm
    real(dp), dimension(max_node_per_cell)      :: duydr, duydm
    real(dp), dimension(max_node_per_cell)      :: duzdr, duzdm
    real(dp), dimension(max_node_per_cell)      :: dvxdr, dvxdn
    real(dp), dimension(max_node_per_cell)      :: dvydr, dvydn
    real(dp), dimension(max_node_per_cell)      :: dvzdr, dvzdn
    real(dp), dimension(max_node_per_cell)      :: dwxdr, dwxdl
    real(dp), dimension(max_node_per_cell)      :: dwydr, dwydl
    real(dp), dimension(max_node_per_cell)      :: dwzdr, dwzdl
    real(dp), dimension(max_node_per_cell)      :: dtxdr, dtxdm, dtxdn
    real(dp), dimension(max_node_per_cell)      :: dtxdl, dtxde
    real(dp), dimension(max_node_per_cell)      :: dtydr, dtydm, dtydn
    real(dp), dimension(max_node_per_cell)      :: dtydl, dtyde
    real(dp), dimension(max_node_per_cell)      :: dtzdr, dtzdm, dtzdn
    real(dp), dimension(max_node_per_cell)      :: dtzdl, dtzde

    real(dp), dimension(max_node_per_cell)      :: dvf2dr, dvf2dm
    real(dp), dimension(max_node_per_cell)      :: dvf2dn, dvf2dl
    real(dp), dimension(max_node_per_cell)      :: dvf3dr, dvf3dm
    real(dp), dimension(max_node_per_cell)      :: dvf3dn, dvf3dl
    real(dp), dimension(max_node_per_cell)      :: dvf4dr, dvf4dm
    real(dp), dimension(max_node_per_cell)      :: dvf4dn, dvf4dl
    real(dp), dimension(max_node_per_cell)      :: dvf5dr, dvf5dm
    real(dp), dimension(max_node_per_cell)      :: dvf5dn, dvf5dl, dvf5de

    real(dp), dimension(max_node_per_cell)      :: x_node, y_node, z_node
    real(dp), dimension(max_node_per_cell)      :: dgradx_celldq, dgrady_celldq
    real(dp), dimension(max_node_per_cell)      :: dgradz_celldq

    real(dp), dimension(max_face_per_cell)      :: nxf, nyf, nzf
    real(dp) :: nx, ny, nz

!   real(dp)      :: mu_node, u_node, v_node, w_node, t_node

    real(dp), dimension(max_node_per_cell)      :: rho_node, e_node, mu_node
    real(dp), dimension(max_node_per_cell)      :: u_node, v_node, w_node
    real(dp), dimension(max_node_per_cell)      :: t_node, p_node

    real(dp), dimension(max_node_per_cell)      :: dudr, dudm
    real(dp), dimension(max_node_per_cell)      :: dvdr, dvdn
    real(dp), dimension(max_node_per_cell)      :: dwdr, dwdl

    real(dp), dimension(17,node_per_cell,node_per_cell) :: a

    real(dp), parameter    :: my_18th = 1.0_dp/18.0_dp
    real(dp), parameter    :: my_4th  = 1.0_dp/4.0_dp
    real(dp), parameter    :: my_6th  = 1.0_dp/6.0_dp
    real(dp), parameter    :: my_3rd  = 1.0_dp/3.0_dp
    real(dp), parameter    :: my_haf  = 0.50_dp
    real(dp), parameter    :: my_mxd  = 0.99939_dp
    real(dp), parameter    :: my_1    = 1.0_dp
    real(dp), parameter    :: my_2    = 2.0_dp
    real(dp), parameter    :: my_3    = 3.0_dp
    real(dp), parameter    :: my_4    = 4.0_dp

    !integer, dimension(2), save :: t_node_loop_1, t_threed_faces, t_jac_convert
    !integer, dimension(2), save :: t_face_angles, t_edge_loop, t_update_loop
    integer, dimension(2), save :: t_update_loop

  continue

    call TAU_PROFILE_INIT()
    call TAU_PROFILE_SET_NODE(0)

    !call TAU_PROFILE_TIMER(t_node_loop_1, 'node_loop_1')
    !call TAU_PROFILE_TIMER(t_threed_faces, 'threed_faces')
    !call TAU_PROFILE_TIMER(t_jac_convert, 'jac_convert')
    !call TAU_PROFILE_TIMER(t_face_angles, 'face_angles')
    !call TAU_PROFILE_TIMER(t_edge_loop, 'edge_loop')
    call TAU_PROFILE_TIMER(t_update_loop, 'update_loop')

    cstar = sutherland_constant/tref
    xmr   = xmach/re
    c43   = my_4/my_3
    c23   = my_2/my_3
    cgp   = xmr/(gm1*prandtl)
    cgpt  = xmr/(gm1*turbulent_prandtl)
    fact = 1.0_dp / real(node_per_cell, dp)

! Loop over the cells in each color

    nstart = 1

    !$omp parallel
    !$&  default(private)
    !$&  firstprivate(nstart)
    !$&  shared(a_diag,a_off,amut,c23,c2n,c43,cgp,cgpt,chk_norm,cstar,
    !$&         edge_per_cell,face_per_cell,fact,g2m,gamma,ggm1,gm1,ia,ivgrd,
    !$&         ja,local_e2n,local_f2n,nnodes0,node_per_cell,
    !$&         nzg2m,qnode,x,xmr,y,z, ncell_in_color,ncell_colors)


    color_loop: do icolor = 1, ncell_colors

    !$omp do schedule(guided)
     cell_loop : do n = nstart, nstart + ncell_in_color(icolor) - 1

! Initialization

      a = 0.0_dp

      rmu    = 0.0_dp
      umu    = 0.0_dp
      vmu    = 0.0_dp
      wmu    = 0.0_dp
      rmucgp = 0.0_dp

      xc = 0._dp
      yc = 0._dp
      zc = 0._dp

! Compute cell averages, cell center, and set up some local solution arrays

      !call TAU_PROFILE_START(t_node_loop_1)

      ! !dir$ ivdep
      node_loop_1 : do i = 1, node_per_cell

        node = c2n(i,n)

        x_node(i) = x(node)
        y_node(i) = y(node)
        z_node(i) = z(node)

        xc = xc + x_node(i)
        yc = yc + y_node(i)
        zc = zc + z_node(i)

        rho_node(i) = qnode(1,node)
        rhoinv = 1.0_dp / rho_node(i)
        u_node(i) = qnode(2,node)*rhoinv
        v_node(i) = qnode(3,node)*rhoinv
        w_node(i) = qnode(4,node)*rhoinv
        e_node(i) = qnode(5,node)
        p_node(i) = gm1*(e_node(i) - my_haf*rho_node(i)*(u_node(i)*u_node(i) + &
                                                         v_node(i)*v_node(i) + &
                                                         w_node(i)*w_node(i)))
        t_node(i)  = gamma*p_node(i)*rhoinv

        mu_node(i) = viscosity_law( cstar, t_node(i) )

        rmucgp = rmucgp + mu_node(i)*cgp + amut(node)*cgpt

        mu_node(i) = mu_node(i) + amut(node)

        rmu = rmu + mu_node(i)
        umu = umu + u_node(i) * mu_node(i)
        vmu = vmu + v_node(i) * mu_node(i)
        wmu = wmu + w_node(i) * mu_node(i)

!     derivatives of (primitive) variables wrt conserved variables
!     for each node in the cell

!     nomenclature:
!       dudr = d(u)/d(rho)
!       dudm = d(u)/d(rho*u)
!       dvdr = d(v)/d(rho)
!       dvdn = d(v)/d(rho*v)
!       dwdr = d(w)/d(rho)
!       dwdl = d(w)/d(rho*w)
!       dtdr = d(t)/d(rho)
!       dtdm = d(t)/d(rho*u)
!       dtdn = d(t)/d(rho*v)
!       dtdl = d(t)/d(rho*w)
!       dtde = d(t)/d(e)

        dudr(i) = - u_node(i) *rhoinv
        dudm(i) =              rhoinv

        dvdr(i) = - v_node(i) *rhoinv
        dvdn(i) =              rhoinv

        dwdr(i) = - w_node(i) *rhoinv
        dwdl(i) =              rhoinv

        dtdr(i) =   ggm1 * ( u_node(i)*u_node(i) +                             &
                             v_node(i)*v_node(i) +                             &
                             w_node(i)*w_node(i) -                             &
                             e_node(i)*rhoinv ) *rhoinv
        dtdm(i) = - ggm1 * u_node(i) *rhoinv
        dtdn(i) = - ggm1 * v_node(i) *rhoinv
        dtdl(i) = - ggm1 * w_node(i) *rhoinv
        dtde(i) =   ggm1 *rhoinv

      end do node_loop_1

      !call TAU_PROFILE_STOP(t_node_loop_1)

!     get cell averages by dividing by the number of nodes that contributed

      rmu    = rmu*fact
      umu    = umu*fact
      vmu    = vmu*fact
      wmu    = wmu*fact
      rmucgp = rmucgp*fact
      xc     = xc*fact
      yc     = yc*fact
      zc     = zc*fact

! Get the jacobians of the gradients in the primal cell via Green-Gauss
! Note: these are with respect to the primitive variables

    dgradx_celldq(:) = 0.0_dp
    dgrady_celldq(:) = 0.0_dp
    dgradz_celldq(:) = 0.0_dp
    cell_vol = 0.0_dp

      !call TAU_PROFILE_START(t_threed_faces)

      threed_faces : do iface = 1,face_per_cell

        nn1 = local_f2n(iface,1)
        nn2 = local_f2n(iface,2)
        nn3 = local_f2n(iface,3)
        nn4 = local_f2n(iface,4)

        if (nn4 == nn1) then

!         triangular faces of the cell

!         face normals (factor of 1/2 deferred till cell_vol
!         and jacobian terms are calculated)

          nx = (y_node(nn2) - y_node(nn1))*(z_node(nn3) - z_node(nn1))         &
             - (z_node(nn2) - z_node(nn1))*(y_node(nn3) - y_node(nn1))
          ny = (z_node(nn2) - z_node(nn1))*(x_node(nn3) - x_node(nn1))         &
             - (x_node(nn2) - x_node(nn1))*(z_node(nn3) - z_node(nn1))
          nz = (x_node(nn2) - x_node(nn1))*(y_node(nn3) - y_node(nn1))         &
             - (y_node(nn2) - y_node(nn1))*(x_node(nn3) - x_node(nn1))

!         face centroid (factor of 1/3 deferred till the
!         contribution to cell_vol is calculated)

          xavg = x_node(nn1) + x_node(nn2) + x_node(nn3)
          yavg = y_node(nn1) + y_node(nn2) + y_node(nn3)
          zavg = z_node(nn1) + z_node(nn2) + z_node(nn3)

!         cell volume contributions

          cell_vol = cell_vol + (xavg*nx + yavg*ny + zavg*nz)

!         jacobian contributions

          dgradx_celldq(nn1) = dgradx_celldq(nn1) + nx
          dgrady_celldq(nn1) = dgrady_celldq(nn1) + ny
          dgradz_celldq(nn1) = dgradz_celldq(nn1) + nz

          dgradx_celldq(nn2) = dgradx_celldq(nn2) + nx
          dgrady_celldq(nn2) = dgrady_celldq(nn2) + ny
          dgradz_celldq(nn2) = dgradz_celldq(nn2) + nz

          dgradx_celldq(nn3) = dgradx_celldq(nn3) + nx
          dgrady_celldq(nn3) = dgrady_celldq(nn3) + ny
          dgradz_celldq(nn3) = dgradz_celldq(nn3) + nz

        else

!         quadrilateral faces of the cell

!         break face up into triangles 1-2-3 and 1-3-4 and add together

!         triangle 1: 1-2-3

!         face centroid (factor of 1/3 deferred till the
!         contribution to cell_vol is calculated)

          xavg1 = x_node(nn1) + x_node(nn2) + x_node(nn3)
          yavg1 = y_node(nn1) + y_node(nn2) + y_node(nn3)
          zavg1 = z_node(nn1) + z_node(nn2) + z_node(nn3)

!         triangle 1 normals (factor of 1/2 deferred till cell_vol
!         and gradient terms are calculated)

          nx1 = (y_node(nn2) - y_node(nn1))*(z_node(nn3) - z_node(nn1))        &
              - (z_node(nn2) - z_node(nn1))*(y_node(nn3) - y_node(nn1))
          ny1 = (z_node(nn2) - z_node(nn1))*(x_node(nn3) - x_node(nn1))        &
              - (x_node(nn2) - x_node(nn1))*(z_node(nn3) - z_node(nn1))
          nz1 = (x_node(nn2) - x_node(nn1))*(y_node(nn3) - y_node(nn1))        &
              - (y_node(nn2) - y_node(nn1))*(x_node(nn3) - x_node(nn1))

          term1 = xavg1*nx1 + yavg1*ny1 + zavg1*nz1

!         triangle 2: 1-3-4

!         face centroid (factor of 1/3 deferred till the
!         contribution to cell_vol is calculated)

          xavg2 = x_node(nn1) + x_node(nn3) + x_node(nn4)
          yavg2 = y_node(nn1) + y_node(nn3) + y_node(nn4)
          zavg2 = z_node(nn1) + z_node(nn3) + z_node(nn4)

!         triangle 2 normals (factor of 1/2 deferred till cell_vol
!         and gradient terms are calculated)

          nx2 = (y_node(nn3) - y_node(nn1))*(z_node(nn4) - z_node(nn1))        &
              - (z_node(nn3) - z_node(nn1))*(y_node(nn4) - y_node(nn1))
          ny2 = (z_node(nn3) - z_node(nn1))*(x_node(nn4) - x_node(nn1))        &
              - (x_node(nn3) - x_node(nn1))*(z_node(nn4) - z_node(nn1))
          nz2 = (x_node(nn3) - x_node(nn1))*(y_node(nn4) - y_node(nn1))        &
              - (y_node(nn3) - y_node(nn1))*(x_node(nn4) - x_node(nn1))

          term2 = xavg2*nx2 + yavg2*ny2 + zavg2*nz2

!         cell volume contributions

          cell_vol = cell_vol + (term1 + term2)

!         jacobian contributions

          dgradx_celldq(nn1) = dgradx_celldq(nn1) + nx1 + nx2
          dgrady_celldq(nn1) = dgrady_celldq(nn1) + ny1 + ny2
          dgradz_celldq(nn1) = dgradz_celldq(nn1) + nz1 + nz2

          dgradx_celldq(nn2) = dgradx_celldq(nn2) + nx1
          dgrady_celldq(nn2) = dgrady_celldq(nn2) + ny1
          dgradz_celldq(nn2) = dgradz_celldq(nn2) + nz1

          dgradx_celldq(nn3) = dgradx_celldq(nn3) + nx1 + nx2
          dgrady_celldq(nn3) = dgrady_celldq(nn3) + ny1 + ny2
          dgradz_celldq(nn3) = dgradz_celldq(nn3) + nz1 + nz2

          dgradx_celldq(nn4) = dgradx_celldq(nn4) + nx2
          dgrady_celldq(nn4) = dgrady_celldq(nn4) + ny2
          dgradz_celldq(nn4) = dgradz_celldq(nn4) + nz2

          nx = nx1 + nx2
          ny = ny1 + ny2
          nz = nz1 + nz2

        end if

        nxf(iface) = nx
        nyf(iface) = ny
        nzf(iface) = nz

      end do threed_faces

      !call TAU_PROFILE_STOP(t_threed_faces)

      !call TAU_PROFILE_START(t_jac_convert)

      cell_vol = cell_vol * my_18th
      cell_vol_inv = my_6th/cell_vol

      dgradx_celldq(:) = dgradx_celldq(:) * cell_vol_inv
      dgrady_celldq(:) = dgrady_celldq(:) * cell_vol_inv
      dgradz_celldq(:) = dgradz_celldq(:) * cell_vol_inv

!     convert to jacobians with respect to conservative variables

      do i = 1, node_per_cell

        duxavgdr(i) = dudr(i)*dgradx_celldq(i)
        duxavgdm(i) = dudm(i)*dgradx_celldq(i)
        dvxavgdr(i) = dvdr(i)*dgradx_celldq(i)
        dvxavgdn(i) = dvdn(i)*dgradx_celldq(i)
        dwxavgdr(i) = dwdr(i)*dgradx_celldq(i)
        dwxavgdl(i) = dwdl(i)*dgradx_celldq(i)
        dtxavgdr(i) = dtdr(i)*dgradx_celldq(i)
        dtxavgdm(i) = dtdm(i)*dgradx_celldq(i)
        dtxavgdn(i) = dtdn(i)*dgradx_celldq(i)
        dtxavgdl(i) = dtdl(i)*dgradx_celldq(i)
        dtxavgde(i) = dtde(i)*dgradx_celldq(i)

        duyavgdr(i) = dudr(i)*dgrady_celldq(i)
        duyavgdm(i) = dudm(i)*dgrady_celldq(i)
        dvyavgdr(i) = dvdr(i)*dgrady_celldq(i)
        dvyavgdn(i) = dvdn(i)*dgrady_celldq(i)
        dwyavgdr(i) = dwdr(i)*dgrady_celldq(i)
        dwyavgdl(i) = dwdl(i)*dgrady_celldq(i)
        dtyavgdr(i) = dtdr(i)*dgrady_celldq(i)
        dtyavgdm(i) = dtdm(i)*dgrady_celldq(i)
        dtyavgdn(i) = dtdn(i)*dgrady_celldq(i)
        dtyavgdl(i) = dtdl(i)*dgrady_celldq(i)
        dtyavgde(i) = dtde(i)*dgrady_celldq(i)

        duzavgdr(i) = dudr(i)*dgradz_celldq(i)
        duzavgdm(i) = dudm(i)*dgradz_celldq(i)
        dvzavgdr(i) = dvdr(i)*dgradz_celldq(i)
        dvzavgdn(i) = dvdn(i)*dgradz_celldq(i)
        dwzavgdr(i) = dwdr(i)*dgradz_celldq(i)
        dwzavgdl(i) = dwdl(i)*dgradz_celldq(i)
        dtzavgdr(i) = dtdr(i)*dgradz_celldq(i)
        dtzavgdm(i) = dtdm(i)*dgradz_celldq(i)
        dtzavgdn(i) = dtdn(i)*dgradz_celldq(i)
        dtzavgdl(i) = dtdl(i)*dgradz_celldq(i)
        dtzavgde(i) = dtde(i)*dgradz_celldq(i)

      end do

      !call TAU_PROFILE_STOP(t_jac_convert)

      !call TAU_PROFILE_START(t_face_angles)
! Check cell-face angles (dot product) - if greater that a specified value,
! we skip the contribution from this cell (currently for 3D cases only)

!     note: only need to check upper part of matrix chk_norm since
!     dot products commute (A*B = B*A) and the diagonal indicates
!     the dot product of a face with itself; also note that nx, ny, nz
!     are not unit normals so we must scale the dot product accordingly

      if ( ivgrd == 1 ) then
        big_angle = 0
        do i=1,face_per_cell
          do j=i+1,face_per_cell
            if (chk_norm(i,j) > 0) then
              dot = nxf(i)*nxf(j) + nyf(i)*nyf(j) + nzf(i)*nzf(j)
!             scale to unit normals
              dot = dot / sqrt(nxf(i)*nxf(i) + nyf(i)*nyf(i) + nzf(i)*nzf(i))
              dot = dot / sqrt(nxf(j)*nxf(j) + nyf(j)*nyf(j) + nzf(j)*nzf(j))
              if (dot >= my_mxd) big_angle = big_angle + 1
            end if
          end do
        end do
        if ( big_angle > 0 )  cycle cell_loop
      end if

      !call TAU_PROFILE_STOP(t_face_angles)

! Next loop over the edges in the cell and get each ones'
! contribution to the jacobian

      !call TAU_PROFILE_START(t_edge_loop)

      edge_loop : do ie = 1, edge_per_cell

!       local node numbers of edge endpoints

        n1_loc = local_e2n(ie,1)
        n2_loc = local_e2n(ie,2)
        n3_loc = local_e2n(ie,3)
        n4_loc = local_e2n(ie,4)
        n5_loc = local_e2n(ie,5)
        n6_loc = local_e2n(ie,6)

! Get this edges' contribution to the dual normal and area

!       edge midpoint

        xm = (x_node(n1_loc) + x_node(n2_loc))*my_haf
        ym = (y_node(n1_loc) + y_node(n2_loc))*my_haf
        zm = (z_node(n1_loc) + z_node(n2_loc))*my_haf

!       compute left face centroid

        if (n4_loc /= 0) then
          xl = (x_node(n1_loc) + x_node(n2_loc) + x_node(n3_loc)               &
              + x_node(n4_loc))*my_4th
          yl = (y_node(n1_loc) + y_node(n2_loc) + y_node(n3_loc)               &
              + y_node(n4_loc))*my_4th
          zl = (z_node(n1_loc) + z_node(n2_loc) + z_node(n3_loc)               &
              + z_node(n4_loc))*my_4th
        else
          xl = (x_node(n1_loc) + x_node(n2_loc) + x_node(n3_loc))*my_3rd
          yl = (y_node(n1_loc) + y_node(n2_loc) + y_node(n3_loc))*my_3rd
          zl = (z_node(n1_loc) + z_node(n2_loc) + z_node(n3_loc))*my_3rd
        end if

!       compute right face centroid

        if (n6_loc /= 0) then
          xr = (x_node(n1_loc) + x_node(n2_loc) + x_node(n5_loc)               &
              + x_node(n6_loc))*my_4th
          yr = (y_node(n1_loc) + y_node(n2_loc) + y_node(n5_loc)               &
              + y_node(n6_loc))*my_4th
          zr = (z_node(n1_loc) + z_node(n2_loc) + z_node(n5_loc)               &
              + z_node(n6_loc))*my_4th
        else
          xr = (x_node(n1_loc) + x_node(n2_loc) + x_node(n5_loc))*my_3rd
          yr = (y_node(n1_loc) + y_node(n2_loc) + y_node(n5_loc))*my_3rd
          zr = (z_node(n1_loc) + z_node(n2_loc) + z_node(n5_loc))*my_3rd
        end if

!       get the contributions to dual normals from the two triangles
!       that form part of the dual-cell surface

!       area of dual triangle xm-xr-xc + area of dual triangle xm-xc-xl

        areax = ((yc-ym)*(zl-zr) - (zc-zm)*(yl-yr))*my_haf
        areay = ((zc-zm)*(xl-xr) - (xc-xm)*(zl-zr))*my_haf
        areaz = ((xc-xm)*(yl-yr) - (yc-ym)*(xl-xr))*my_haf

! Get (jacobians of) gradients at the dual face; either take gradients for
! this piece of the dual face to be the same as the cell-average gradient
! computed above  (which is what the legacy FUN3D solver does for tets),
! or combine with the edge-gradient to increase h-ellipticity on
! non-simplicial meshes.

! For tets in 3D or prisms in 2D, edge gradients add no new info
! so there is no need to do the extra work

!         ex, ey, ez is unit vector along edge direction

          ex   = x_node(n2_loc) - x_node(n1_loc)
          ey   = y_node(n2_loc) - y_node(n1_loc)
          ez   = z_node(n2_loc) - z_node(n1_loc)
          disi = my_1/sqrt( ex**2 + ey**2 + ez**2 )

          ex   = ex*disi
          ey   = ey*disi
          ez   = ez*disi

!           write the above gradients as
!           ux = uxavg - (uxavg*ex + uyavg*ey + uzavg*ez)*ex + (u2-u1)*disi*ex
!                <-----------------avg_term---------------->   <--edge_term-->
!           etc.

!           first get the avg_term pieces; all active cell nodes contribute

            do i = 1, node_per_cell

!             u-gradient avg terms

              dgradu_xidr = duxavgdr(i)*ex + duyavgdr(i)*ey +  duzavgdr(i)*ez
              dgradu_xidm = duxavgdm(i)*ex + duyavgdm(i)*ey +  duzavgdm(i)*ez

              duxdr(i) = duxavgdr(i) - dgradu_xidr*ex
              duxdm(i) = duxavgdm(i) - dgradu_xidm*ex
              duydr(i) = duyavgdr(i) - dgradu_xidr*ey
              duydm(i) = duyavgdm(i) - dgradu_xidm*ey
              duzdr(i) = duzavgdr(i) - dgradu_xidr*ez
              duzdm(i) = duzavgdm(i) - dgradu_xidm*ez

!             v-gradient avg terms

              dgradv_xidr = dvxavgdr(i)*ex + dvyavgdr(i)*ey +  dvzavgdr(i)*ez
              dgradv_xidn = dvxavgdn(i)*ex + dvyavgdn(i)*ey +  dvzavgdn(i)*ez

              dvxdr(i) = dvxavgdr(i) - dgradv_xidr*ex
              dvxdn(i) = dvxavgdn(i) - dgradv_xidn*ex
              dvydr(i) = dvyavgdr(i) - dgradv_xidr*ey
              dvydn(i) = dvyavgdn(i) - dgradv_xidn*ey
              dvzdr(i) = dvzavgdr(i) - dgradv_xidr*ez
              dvzdn(i) = dvzavgdn(i) - dgradv_xidn*ez

!             w-gradient avg terms

              dgradw_xidr = dwxavgdr(i)*ex + dwyavgdr(i)*ey +  dwzavgdr(i)*ez
              dgradw_xidl = dwxavgdl(i)*ex + dwyavgdl(i)*ey +  dwzavgdl(i)*ez

              dwxdr(i) = dwxavgdr(i) - dgradw_xidr*ex
              dwxdl(i) = dwxavgdl(i) - dgradw_xidl*ex
              dwydr(i) = dwyavgdr(i) - dgradw_xidr*ey
              dwydl(i) = dwyavgdl(i) - dgradw_xidl*ey
              dwzdr(i) = dwzavgdr(i) - dgradw_xidr*ez
              dwzdl(i) = dwzavgdl(i) - dgradw_xidl*ez

!             t-gradient avg terms

              dgradt_xidr = dtxavgdr(i)*ex + dtyavgdr(i)*ey +  dtzavgdr(i)*ez
              dgradt_xidm = dtxavgdm(i)*ex + dtyavgdm(i)*ey +  dtzavgdm(i)*ez
              dgradt_xidn = dtxavgdn(i)*ex + dtyavgdn(i)*ey +  dtzavgdn(i)*ez
              dgradt_xidl = dtxavgdl(i)*ex + dtyavgdl(i)*ey +  dtzavgdl(i)*ez
              dgradt_xide = dtxavgde(i)*ex + dtyavgde(i)*ey +  dtzavgde(i)*ez

              dtxdr(i) = dtxavgdr(i) - dgradt_xidr*ex
              dtxdm(i) = dtxavgdm(i) - dgradt_xidm*ex
              dtxdn(i) = dtxavgdn(i) - dgradt_xidn*ex
              dtxdl(i) = dtxavgdl(i) - dgradt_xidl*ex
              dtxde(i) = dtxavgde(i) - dgradt_xide*ex

              dtydr(i) = dtyavgdr(i) - dgradt_xidr*ey
              dtydm(i) = dtyavgdm(i) - dgradt_xidm*ey
              dtydn(i) = dtyavgdn(i) - dgradt_xidn*ey
              dtydl(i) = dtyavgdl(i) - dgradt_xidl*ey
              dtyde(i) = dtyavgde(i) - dgradt_xide*ey

              dtzdr(i) = dtzavgdr(i) - dgradt_xidr*ez
              dtzdm(i) = dtzavgdm(i) - dgradt_xidm*ez
              dtzdn(i) = dtzavgdn(i) - dgradt_xidn*ez
              dtzdl(i) = dtzavgdl(i) - dgradt_xidl*ez
              dtzde(i) = dtzavgde(i) - dgradt_xide*ez

            end do

!           next get the edge_term pieces; only the two edge nodes contribute
!           u-gradient edge terms

            disiex = disi*ex
            disiey = disi*ey
            disiez = disi*ez

            duxdr(n1_loc) = duxdr(n1_loc) - dudr(n1_loc)*disiex
            duxdr(n2_loc) = duxdr(n2_loc) + dudr(n2_loc)*disiex
            duxdm(n1_loc) = duxdm(n1_loc) - dudm(n1_loc)*disiex
            duxdm(n2_loc) = duxdm(n2_loc) + dudm(n2_loc)*disiex
            duydr(n1_loc) = duydr(n1_loc) - dudr(n1_loc)*disiey
            duydr(n2_loc) = duydr(n2_loc) + dudr(n2_loc)*disiey
            duydm(n1_loc) = duydm(n1_loc) - dudm(n1_loc)*disiey
            duydm(n2_loc) = duydm(n2_loc) + dudm(n2_loc)*disiey
            duzdr(n1_loc) = duzdr(n1_loc) - dudr(n1_loc)*disiez
            duzdr(n2_loc) = duzdr(n2_loc) + dudr(n2_loc)*disiez
            duzdm(n1_loc) = duzdm(n1_loc) - dudm(n1_loc)*disiez
            duzdm(n2_loc) = duzdm(n2_loc) + dudm(n2_loc)*disiez

!           v-gradient edge terms

            dvxdr(n1_loc) = dvxdr(n1_loc) - dvdr(n1_loc)*disiex
            dvxdr(n2_loc) = dvxdr(n2_loc) + dvdr(n2_loc)*disiex
            dvxdn(n1_loc) = dvxdn(n1_loc) - dvdn(n1_loc)*disiex
            dvxdn(n2_loc) = dvxdn(n2_loc) + dvdn(n2_loc)*disiex
            dvydr(n1_loc) = dvydr(n1_loc) - dvdr(n1_loc)*disiey
            dvydr(n2_loc) = dvydr(n2_loc) + dvdr(n2_loc)*disiey
            dvydn(n1_loc) = dvydn(n1_loc) - dvdn(n1_loc)*disiey
            dvydn(n2_loc) = dvydn(n2_loc) + dvdn(n2_loc)*disiey
            dvzdr(n1_loc) = dvzdr(n1_loc) - dvdr(n1_loc)*disiez
            dvzdr(n2_loc) = dvzdr(n2_loc) + dvdr(n2_loc)*disiez
            dvzdn(n1_loc) = dvzdn(n1_loc) - dvdn(n1_loc)*disiez
            dvzdn(n2_loc) = dvzdn(n2_loc) + dvdn(n2_loc)*disiez

!           w-gradient edge terms

            dwxdr(n1_loc) = dwxdr(n1_loc) - dwdr(n1_loc)*disiex
            dwxdr(n2_loc) = dwxdr(n2_loc) + dwdr(n2_loc)*disiex
            dwxdl(n1_loc) = dwxdl(n1_loc) - dwdl(n1_loc)*disiex
            dwxdl(n2_loc) = dwxdl(n2_loc) + dwdl(n2_loc)*disiex
            dwydr(n1_loc) = dwydr(n1_loc) - dwdr(n1_loc)*disiey
            dwydr(n2_loc) = dwydr(n2_loc) + dwdr(n2_loc)*disiey
            dwydl(n1_loc) = dwydl(n1_loc) - dwdl(n1_loc)*disiey
            dwydl(n2_loc) = dwydl(n2_loc) + dwdl(n2_loc)*disiey
            dwzdr(n1_loc) = dwzdr(n1_loc) - dwdr(n1_loc)*disiez
            dwzdr(n2_loc) = dwzdr(n2_loc) + dwdr(n2_loc)*disiez
            dwzdl(n1_loc) = dwzdl(n1_loc) - dwdl(n1_loc)*disiez
            dwzdl(n2_loc) = dwzdl(n2_loc) + dwdl(n2_loc)*disiez

!           t-gradient edge terms

            dtxdr(n1_loc) = dtxdr(n1_loc) - dtdr(n1_loc)*disiex
            dtxdr(n2_loc) = dtxdr(n2_loc) + dtdr(n2_loc)*disiex
            dtxdm(n1_loc) = dtxdm(n1_loc) - dtdm(n1_loc)*disiex
            dtxdm(n2_loc) = dtxdm(n2_loc) + dtdm(n2_loc)*disiex
            dtxdn(n1_loc) = dtxdn(n1_loc) - dtdn(n1_loc)*disiex
            dtxdn(n2_loc) = dtxdn(n2_loc) + dtdn(n2_loc)*disiex
            dtxdl(n1_loc) = dtxdl(n1_loc) - dtdl(n1_loc)*disiex
            dtxdl(n2_loc) = dtxdl(n2_loc) + dtdl(n2_loc)*disiex
            dtxde(n1_loc) = dtxde(n1_loc) - dtde(n1_loc)*disiex
            dtxde(n2_loc) = dtxde(n2_loc) + dtde(n2_loc)*disiex
            dtydr(n1_loc) = dtydr(n1_loc) - dtdr(n1_loc)*disiey
            dtydr(n2_loc) = dtydr(n2_loc) + dtdr(n2_loc)*disiey
            dtydm(n1_loc) = dtydm(n1_loc) - dtdm(n1_loc)*disiey
            dtydm(n2_loc) = dtydm(n2_loc) + dtdm(n2_loc)*disiey
            dtydn(n1_loc) = dtydn(n1_loc) - dtdn(n1_loc)*disiey
            dtydn(n2_loc) = dtydn(n2_loc) + dtdn(n2_loc)*disiey
            dtydl(n1_loc) = dtydl(n1_loc) - dtdl(n1_loc)*disiey
            dtydl(n2_loc) = dtydl(n2_loc) + dtdl(n2_loc)*disiey
            dtyde(n1_loc) = dtyde(n1_loc) - dtde(n1_loc)*disiey
            dtyde(n2_loc) = dtyde(n2_loc) + dtde(n2_loc)*disiey
            dtzdr(n1_loc) = dtzdr(n1_loc) - dtdr(n1_loc)*disiez
            dtzdr(n2_loc) = dtzdr(n2_loc) + dtdr(n2_loc)*disiez
            dtzdm(n1_loc) = dtzdm(n1_loc) - dtdm(n1_loc)*disiez
            dtzdm(n2_loc) = dtzdm(n2_loc) + dtdm(n2_loc)*disiez
            dtzdn(n1_loc) = dtzdn(n1_loc) - dtdn(n1_loc)*disiez
            dtzdn(n2_loc) = dtzdn(n2_loc) + dtdn(n2_loc)*disiez
            dtzdl(n1_loc) = dtzdl(n1_loc) - dtdl(n1_loc)*disiez
            dtzdl(n2_loc) = dtzdl(n2_loc) + dtdl(n2_loc)*disiez
            dtzde(n1_loc) = dtzde(n1_loc) - dtde(n1_loc)*disiez
            dtzde(n2_loc) = dtzde(n2_loc) + dtde(n2_loc)*disiez


! Viscous contributions at dual face [ full Navier-Stokes terms ]

!       [nondimensionalization factor xmr ] * [ viscosity ]
!       [ unit normal and area ] at dual face

        rax = xmr*rmu*areax
        ray = xmr*rmu*areay
        raz = xmr*rmu*areaz

        tax = rmucgp*areax
        tay = rmucgp*areay
        taz = rmucgp*areaz

!       form some intermediate Jacobians at all nodes

        rmuinv = 1.0_dp/rmu

        do i = 1, node_per_cell

! Stress pieces

          dtxxdr =   c43*duxdr(i) - c23*dvydr(i) - c23*dwzdr(i)
          dtxxdm =   c43*duxdm(i)
          dtxxdn = - c23*dvydn(i)
          dtxxdl = - c23*dwzdl(i)

          dtxydr = duydr(i) + dvxdr(i)
          dtxydm = duydm(i)
          dtxydn = dvxdn(i)

          dtxzdr = duzdr(i) + dwxdr(i)
          dtxzdm = duzdm(i)
          dtxzdl = dwxdl(i)

          dtyydr =   c43*dvydr(i) - c23*duxdr(i) - c23*dwzdr(i)
          dtyydm = - c23*duxdm(i)
          dtyydn =   c43*dvydn(i)
          dtyydl = - c23*dwzdl(i)

          dtyzdr = dvzdr(i) + dwydr(i)
          dtyzdn = dvzdn(i)
          dtyzdl = dwydl(i)

          dtzzdr =   c43*dwzdr(i) - c23*duxdr(i) - c23*dvydr(i)
          dtzzdm = - c23*duxdm(i)
          dtzzdn = - c23*dvydn(i)
          dtzzdl =   c43*dwzdl(i)

! Add stress contributions

          dvf2dr(i) = rax*dtxxdr + ray*dtxydr + raz*dtxzdr
          dvf2dm(i) = rax*dtxxdm + ray*dtxydm + raz*dtxzdm
          dvf2dn(i) = rax*dtxxdn + ray*dtxydn
          dvf2dl(i) = rax*dtxxdl              + raz*dtxzdl

          dvf3dr(i) = rax*dtxydr + ray*dtyydr + raz*dtyzdr
          dvf3dm(i) = rax*dtxydm + ray*dtyydm
          dvf3dn(i) = rax*dtxydn + ray*dtyydn + raz*dtyzdn
          dvf3dl(i) =              ray*dtyydl + raz*dtyzdl

          dvf4dr(i) = rax*dtxzdr + ray*dtyzdr + raz*dtzzdr
          dvf4dm(i) = rax*dtxzdm              + raz*dtzzdm
          dvf4dn(i) =              ray*dtyzdn + raz*dtzzdn
          dvf4dl(i) = rax*dtxzdl + ray*dtyzdl + raz*dtzzdl

          dvf5dr(i) = tax*dtxdr(i) + tay*dtydr(i) + taz*dtzdr(i)             &
                    + (umu*dvf2dr(i) + vmu*dvf3dr(i) + wmu*dvf4dr(i))*rmuinv
          dvf5dm(i) = tax*dtxdm(i) + tay*dtydm(i) + taz*dtzdm(i)             &
                    + (umu*dvf2dm(i) + vmu*dvf3dm(i) + wmu*dvf4dm(i))*rmuinv
          dvf5dn(i) = tax*dtxdn(i) + tay*dtydn(i) + taz*dtzdn(i)             &
                    + (umu*dvf2dn(i) + vmu*dvf3dn(i) + wmu*dvf4dn(i))*rmuinv
          dvf5dl(i) = tax*dtxdl(i) + tay*dtydl(i) + taz*dtzdl(i)             &
                    + (umu*dvf2dl(i) + vmu*dvf3dl(i) + wmu*dvf4dl(i))*rmuinv
          dvf5de(i) = tax*dtxde(i) + tay*dtyde(i) + taz*dtzde(i)

        end do

! Assemble final Jacobian matrices into sparse matrix form
! Map the local entries into a compact 17 spaces (instead of 25)
! and unpack later

          node_loop_3 : do j = 1, node_per_cell

            a(1,n1_loc,j)  = a(1,n1_loc,j)  - dvf2dr(j)
            a(2,n1_loc,j)  = a(2,n1_loc,j)  - dvf3dr(j)
            a(3,n1_loc,j)  = a(3,n1_loc,j)  - dvf4dr(j)
            a(4,n1_loc,j)  = a(4,n1_loc,j)  - dvf5dr(j)

            a(5,n1_loc,j)  = a(5,n1_loc,j)  - dvf2dm(j)
            a(6,n1_loc,j)  = a(6,n1_loc,j)  - dvf3dm(j)
            a(7,n1_loc,j)  = a(7,n1_loc,j)  - dvf4dm(j)
            a(8,n1_loc,j)  = a(8,n1_loc,j)  - dvf5dm(j)

            a(9,n1_loc,j)  = a(9,n1_loc,j)  - dvf2dn(j)
            a(10,n1_loc,j) = a(10,n1_loc,j) - dvf3dn(j)
            a(11,n1_loc,j) = a(11,n1_loc,j) - dvf4dn(j)
            a(12,n1_loc,j) = a(12,n1_loc,j) - dvf5dn(j)

            a(13,n1_loc,j) = a(13,n1_loc,j) - dvf2dl(j)
            a(14,n1_loc,j) = a(14,n1_loc,j) - dvf3dl(j)
            a(15,n1_loc,j) = a(15,n1_loc,j) - dvf4dl(j)
            a(16,n1_loc,j) = a(16,n1_loc,j) - dvf5dl(j)

            a(17,n1_loc,j) = a(17,n1_loc,j) - dvf5de(j)

          end do node_loop_3

          node_loop_4 : do j = 1, node_per_cell

            a(1,n2_loc,j)  = a(1,n2_loc,j)  + dvf2dr(j)
            a(2,n2_loc,j)  = a(2,n2_loc,j)  + dvf3dr(j)
            a(3,n2_loc,j)  = a(3,n2_loc,j)  + dvf4dr(j)
            a(4,n2_loc,j)  = a(4,n2_loc,j)  + dvf5dr(j)

            a(5,n2_loc,j)  = a(5,n2_loc,j)  + dvf2dm(j)
            a(6,n2_loc,j)  = a(6,n2_loc,j)  + dvf3dm(j)
            a(7,n2_loc,j)  = a(7,n2_loc,j)  + dvf4dm(j)
            a(8,n2_loc,j)  = a(8,n2_loc,j)  + dvf5dm(j)

            a(9,n2_loc,j)  = a(9,n2_loc,j)  + dvf2dn(j)
            a(10,n2_loc,j) = a(10,n2_loc,j) + dvf3dn(j)
            a(11,n2_loc,j) = a(11,n2_loc,j) + dvf4dn(j)
            a(12,n2_loc,j) = a(12,n2_loc,j) + dvf5dn(j)

            a(13,n2_loc,j) = a(13,n2_loc,j) + dvf2dl(j)
            a(14,n2_loc,j) = a(14,n2_loc,j) + dvf3dl(j)
            a(15,n2_loc,j) = a(15,n2_loc,j) + dvf4dl(j)
            a(16,n2_loc,j) = a(16,n2_loc,j) + dvf5dl(j)

            a(17,n2_loc,j) = a(17,n2_loc,j) + dvf5de(j)

          end do node_loop_4

      end do edge_loop

      !call TAU_PROFILE_STOP(t_edge_loop)

      call TAU_PROFILE_START(t_update_loop)

      do i = 1, node_per_cell
        node = c2n(i,n)
        if ( node <= nnodes0 ) then

          do j = 1, node_per_cell

            if ( i == j ) then

              row = g2m(node)

              a_diag(2,1,row) = a_diag(2,1,row) + a(1,i,i)
              a_diag(3,1,row) = a_diag(3,1,row) + a(2,i,i)
              a_diag(4,1,row) = a_diag(4,1,row) + a(3,i,i)
              a_diag(5,1,row) = a_diag(5,1,row) + a(4,i,i)

              a_diag(2,2,row) = a_diag(2,2,row) + a(5,i,i)
              a_diag(3,2,row) = a_diag(3,2,row) + a(6,i,i)
              a_diag(4,2,row) = a_diag(4,2,row) + a(7,i,i)
              a_diag(5,2,row) = a_diag(5,2,row) + a(8,i,i)

              a_diag(2,3,row) = a_diag(2,3,row) + a(9,i,i)
              a_diag(3,3,row) = a_diag(3,3,row) + a(10,i,i)
              a_diag(4,3,row) = a_diag(4,3,row) + a(11,i,i)
              a_diag(5,3,row) = a_diag(5,3,row) + a(12,i,i)

              a_diag(2,4,row) = a_diag(2,4,row) + a(13,i,i)
              a_diag(3,4,row) = a_diag(3,4,row) + a(14,i,i)
              a_diag(4,4,row) = a_diag(4,4,row) + a(15,i,i)
              a_diag(5,4,row) = a_diag(5,4,row) + a(16,i,i)

              a_diag(5,5,row) = a_diag(5,5,row) + a(17,i,i)

            else

              nodec = c2n(j,n)

              ioff = 0
              search : do k = ia(node), ia(node+1) - 1
                column = ja(k)
                if (column == nodec) then
                  ioff = nzg2m(k)
                  exit search
                endif
              end do search

              a_off(2,1,ioff) = a_off(2,1,ioff) + real(a(1,i,j),odp)
              a_off(3,1,ioff) = a_off(3,1,ioff) + real(a(2,i,j),odp)
              a_off(4,1,ioff) = a_off(4,1,ioff) + real(a(3,i,j),odp)
              a_off(5,1,ioff) = a_off(5,1,ioff) + real(a(4,i,j),odp)

              a_off(2,2,ioff) = a_off(2,2,ioff) + real(a(5,i,j),odp)
              a_off(3,2,ioff) = a_off(3,2,ioff) + real(a(6,i,j),odp)
              a_off(4,2,ioff) = a_off(4,2,ioff) + real(a(7,i,j),odp)
              a_off(5,2,ioff) = a_off(5,2,ioff) + real(a(8,i,j),odp)

              a_off(2,3,ioff) = a_off(2,3,ioff) + real(a(9,i,j),odp)
              a_off(3,3,ioff) = a_off(3,3,ioff) + real(a(10,i,j),odp)
              a_off(4,3,ioff) = a_off(4,3,ioff) + real(a(11,i,j),odp)
              a_off(5,3,ioff) = a_off(5,3,ioff) + real(a(12,i,j),odp)

              a_off(2,4,ioff) = a_off(2,4,ioff) + real(a(13,i,j),odp)
              a_off(3,4,ioff) = a_off(3,4,ioff) + real(a(14,i,j),odp)
              a_off(4,4,ioff) = a_off(4,4,ioff) + real(a(15,i,j),odp)
              a_off(5,4,ioff) = a_off(5,4,ioff) + real(a(16,i,j),odp)

              a_off(5,5,ioff) = a_off(5,5,ioff) + real(a(17,i,j),odp)

            endif
          end do

        end if
      end do

      call TAU_PROFILE_STOP(t_update_loop)

    end do cell_loop
    !$omp end do

     nstart = nstart + ncell_in_color(icolor)

    end do color_loop
    !$omp end parallel

  end subroutine edgejp_mix_cell_opt

  subroutine edgejp_done
    write(*,*) 'edgejp done'
  end subroutine edgejp_done

end module jacobian_module


program main

  use omp_lib,         only : omp_get_num_threads, omp_get_thread_num, &
       omp_get_wtime
  use kinddefs,        only : dp
  use jacobian_module, only : edgejp_mix_cell_opt, edgejp_done
  use ieee_arithmetic, only : ieee_is_nan
  use data_module,     only : load_data, golden, datafile, timing, a_diag,     &
                              a_off, color_flag, color_cells, ncell_colors,    &
                              ncell_in_color

  implicit none

  integer :: i, j, k

  real(dp) :: rms, t1, t2

  !real(dp), parameter :: tol = 2.e-06_dp
  real(dp), parameter :: tol = 1.e-05_dp

  call get_command_argument(1,value=datafile)
  if ( trim(datafile) == '' ) then
    write(*,*) 'Error reading datafile name from command line.'
    stop
  endif

  call get_command_argument(2,value=color_flag)
  if ( trim(color_flag) == '' ) then
    write(*,*) 'Error reading coloring flag from command line.'
    stop
  endif

! Load the dataset for testing

  call load_data()

! Color the cells (or not)

  call color_cells()

  write(*,*) 'Number of colors = ', ncell_colors

  write(*,*) 'Number of cells in each color = ', ncell_in_color(:)

#ifdef _OPENMP
  t1 = omp_get_wtime()
#else
  call cpu_time(t1)
#endif

! Linearize the fluxes

  write(*,*) 'Linearizing viscous flux...'
#ifdef _OPENMP
  !$omp parallel
  if (omp_get_thread_num() == 0) &
       write(*,*) 'on ', omp_get_num_threads(), ' threads'
  !$omp end parallel
#endif

  call edgejp_mix_cell_opt()

  call edgejp_done()

#ifdef _OPENMP
  t2 = omp_get_wtime()
#else
  call cpu_time(t2)
#endif

  if ( timing ) write(*,*) 'Total time taken = ', t2-t1

! Test to see if answer is correct

  rms = 0.0_dp

  do i = 1, size(a_diag,3)
    do j = 1, size(a_diag,1)
      do k = 1, size(a_diag,2)
        rms = rms + a_diag(k,j,i)**2
      end do
    end do
  end do

  do i = 1, size(a_off,3)
    do j = 1, size(a_off,1)
      do k = 1, size(a_off,2)
        rms = rms + a_off(k,j,i)**2
      end do
    end do
  end do

  rms = sqrt(rms)

  if ( ieee_is_nan(rms) .or. abs(rms-golden) >= tol ) then
    write(*,*) 'TEST FAILED:', rms, golden
  else
    write(*,*) 'Test passed.', rms, golden
  endif

end program main

